/* 
	ctrlproxy: A modular IRC proxy
	(c) 2002 Jelmer Vernooij <jelmer@nl.linux.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _GNU_SOURCE
#include "ctrlproxy.h"
#include <string.h>

static time_t last_message = 0, max_idle_time = 60 * 10;

static gboolean is_away = FALSE;

static char *message = NULL;

gboolean check_time(gpointer user_data) {
	if(time(NULL) - last_message > max_idle_time && !is_away) { 
		GList *sl = servers;
		is_away = TRUE;
		while(sl) {
			struct network *s = (struct network *)sl->data;
			irc_send_args(s->outgoing, "AWAY", message?message:"Auto Away", NULL);
			sl = g_list_next(sl);
		}
	}

	return TRUE;
}

static gboolean log_data(struct line *l) {
	if(l->direction == TO_SERVER && !strcasecmp(l->args[0], "AWAY")) {
		if(l->args[1])is_away = TRUE;
		else is_away = FALSE;
	}

	if(l->direction == TO_SERVER &&  
	   (!strcasecmp(l->args[0], "PRIVMSG") || !strcasecmp(l->args[0], "NOTICE"))) {
		last_message = time(NULL);
		if(is_away) {
			GList *sl = servers;
			while(sl) {
				struct network *s = (struct network *)sl->data;
				irc_send_args(s->outgoing, "AWAY", NULL);
				sl = g_list_next(sl);
			}
			is_away = FALSE;
		}
	}
	return TRUE;
}

void init_plugin(GModule *g, xmlNodePtr p) {
	GHook *h;
	char *t = NULL;
	xmlNodePtr cur = p->xmlChildrenNode;
	h = g_hook_alloc(&data_hook);
	h->func = log_data;
	g_hook_append(&data_hook, h);
	last_message = time(NULL);
	
	while(cur) {
		if(!xmlIsBlankNode(cur) && !strcmp(cur->name, "message")) { 
			message = xmlNodeGetContent(cur); 
			t = xmlGetProp(cur, "time"); 
		}
		cur = cur->next;
	}

	if(t) {
		max_idle_time = atol(t);
		xmlFree(t);
	}

	g_timeout_add(1000, check_time, NULL);
}
