/* 
	ctrlproxy: A modular IRC proxy
	(c) 2002 Jelmer Vernooij <jelmer@nl.linux.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _GNU_SOURCE
#include "ctrlproxy.h"
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <time.h>
#include <glib.h>

#undef G_LOG_DOMAIN
#define G_LOG_DOMAIN "log_irssi"

static char *logfile = NULL;
static GHashTable *files = NULL;


static FILE *find_channel_file(struct network *s, char *name) {
	char *n = NULL;
	FILE *f;
	f = g_hash_table_lookup(files, (!name?"messages":name));
	if(!f) {
		char *cn, *server_name;

		server_name = xmlGetProp(s->xmlConf, "name");

		if(strchr(server_name, '/'))server_name = strrchr(server_name, '/');

		asprintf(&n, "%s/%s", logfile, server_name);
		/* Check if directory needs to be created */
		if(access(n, F_OK) != 0 && mkdir(n, 0700) == -1) {
			g_warning("Couldn't create directory %s for logging!", n);
			free(n);
			xmlFree(server_name);
			return NULL;
		}
		free(n);
		
		/* Then open the correct filename */
		cn = g_ascii_strdown(name?name:"messages", -1);
		asprintf(&n, "%s/%s/%s", logfile, server_name, cn);
		xmlFree(server_name);
		g_free(cn);
		f = fopen(n, "a+");
		if(!f) {
			g_warning("Couldn't open file %s for logging!", n);
			free(n);
			return NULL;
		}
		free(n);
		g_hash_table_insert(files, strdup(name), f);
	}
	g_assert(f);
	return f;
}

static gboolean log_data(struct line *l)
{
	char *nick = NULL;
	char *dest = NULL;
	char *user = NULL;
	time_t ti = time(NULL);
	struct tm *t = localtime(&ti);
	FILE *f = NULL;
	if(!l->args || !l->args[0])return TRUE;
	if(l->origin)nick = strdup(l->origin);
	if(nick)user = strchr(nick, '!');
	if(user){ *user = '\0';user++; }
	if(!nick && xmlHasProp(l->network->xmlConf, "nick"))nick = xmlGetProp(l->network->xmlConf, "nick");

	g_assert(l->args[0]);

	if(l->direction == FROM_SERVER && !strcasecmp(l->args[0], "JOIN")) {
		f = find_channel_file(l->network, l->args[1]);
		if(f)fprintf(f, "%02d:%02d -!- %s [%s] has joined %s\n", t->tm_hour, t->tm_min, nick, user, l->args[1]);
	} else if(l->direction == FROM_SERVER && !strcasecmp(l->args[0], "PART")) {
		f = find_channel_file(l->network, l->args[1]);
		if(f)fprintf(f, "%02d:%02d -!- %s [%s] has left %s [%s]\n", t->tm_hour, t->tm_min, nick, user, l->args[1], l->args[2]?l->args[2]:"");
	} else if(!strcasecmp(l->args[0], "PRIVMSG")) {
		dest = l->args[1];
		char *nnick = xmlGetProp(l->network->xmlConf, "nick");
		if(!strcasecmp(dest, nnick))dest = nick;
		xmlFree(nnick);
		if(l->args[2][0] == '') { 
			l->args[2][strlen(l->args[2])-1] = '\0';
			if(!strncasecmp(l->args[2], "ACTION ", 8)) { 
				f = find_channel_file(l->network, dest);
				if(f)fprintf(f, "%02d:%02d  * %s %s\n", t->tm_hour, t->tm_min, nick, l->args[2]+8);
			}
			l->args[2][strlen(l->args[2])] = '';
			/* Ignore all other ctcp messages */
		} else {
			f = find_channel_file(l->network, dest);
			if(f)fprintf(f, "%02d:%02d < %s> %s\n", t->tm_hour, t->tm_min, nick, l->args[2]);
		}
	} else if(!strcasecmp(l->args[0], "MODE") && (l->args[1][1] == '#' || l->args[1][1] == '&')) {
		f = find_channel_file(l->network, l->args[1]);
		if(f)fprintf(f, "%02d:%02d -!- mode/%s [%s %s] by %s\n", t->tm_hour, t->tm_min, l->args[1], l->args[2], l->args[3], nick);
	}

	if(f)fflush(f);

	if(nick)free(nick);

	return TRUE;
}

void init_plugin(GModule *m, xmlNodePtr p)
{
	GHook *h;
	xmlNodePtr cur = p->xmlChildrenNode;
	while(cur) {
		if(!xmlIsBlankNode(cur) && !strcmp(cur->name, "logfile")) logfile = xmlNodeGetContent(cur);
		cur = cur->next;
	}
	g_assert(logfile);
	if(logfile) {
		files = g_hash_table_new(g_str_hash, g_str_equal);
		h = g_hook_alloc(&data_hook);
		h->func = log_data;
		g_hook_append(&data_hook, h);
	}
}
