/* 
	ctrlproxy: A modular IRC proxy
	(c) 2002 Jelmer Vernooij <jelmer@nl.linux.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define _GNU_SOURCE
#include "ctrlproxy.h"
#include <string.h>

CTRLPROXY_VERSIONING_MAGIC

static GHashTable *files = NULL;

void replicate_data(struct server *s, GIOChannel *io) {
	GList *l = g_hash_table_lookup(files, s);
	if(!l){  default_replicate_function(s, io); return; }
	g_hash_table_remove(files, s);
}

static gboolean log_data(struct line *l) {
	GIOChannel *gl = g_hash_table_lookup(files, l->server);
	struct channel *c;

	if(l->argc < 1)return TRUE;
	
	if(l->direction == TO_SERVER &&  
	   (!strcasecmp(l->args[0], "PRIVMSG") || !strcasecmp(l->args[0], "NOTICE"))) {
		g_list_free(gl);
		gl = NULL;
		g_hash_table_replace(files, l->server, NULL);
		return TRUE;
	}

	if(l->direction == TO_SERVER)return TRUE;

	if(!gl) { 
		GList *lg = gen_replication(l->server);
		gl = g_io_channel_
	}
	

	if(!strcasecmp(l->args[0], "PRIVMSG") ||
	   !strcasecmp(l->args[0], "NOTICE") ||
	   !strcasecmp(l->args[0], "MODE") || 
	   !strcasecmp(l->args[0], "JOIN") || 
	   !strcasecmp(l->args[0], "KICK") || 
	   !strcasecmp(l->args[0], "QUIT") ||
	   !strcasecmp(l->args[0], "TOPIC") ||
	   !strcasecmp(l->args[0], "NICK")) {
		
		gl = g_list_append(gl, linedup(l));
	} else if(!strcasecmp(l->args[0], "353")) {
		c = find_channel(l->server, l->args[3]);
		if(c && !(c->introduced & 2)) {
			gl = g_list_append(gl, linedup(l));
		}
		/* Only do 366 if not & 2. Set | 2 */
	} else if(!strcasecmp(l->args[0], "366")) {
		c = find_channel(l->server, l->args[2]);
		if(c && !(c->introduced & 2)) {
			gl = g_list_append(gl, linedup(l));
			c->introduced |= 2;
		}
		/* Only do 331 or 332 if not & 1. Set | 1 */
	} else if(!strcasecmp(l->args[0], "331") || !strcasecmp(l->args[0], "332")) {
		c = find_channel(l->server, l->args[2]);
		if(c && !(c->introduced & 1)) {
			gl = g_list_append(gl, linedup(l));
			c->introduced |= 1;
		}
	}

	g_hash_table_replace(files, l->server, gl);
	return TRUE;
}

void init_plugin(GModule *g, xmlNodePtr p) {
	GHook *h;
	files = g_hash_table_new(NULL, NULL);
	replicate_function = replicate_data;
	h = g_hook_alloc(&data_hook);
	h->func = log_data;
	g_hook_append(&data_hook, h);
}
