/* 
	ctrlproxy: A modular IRC proxy
	(c) 2002 Jelmer Vernooij <jelmer@nl.linux.org>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "internals.h"
#include "config.h"

#define SERVER_SEND_LINE(s,l) {if(!irc_send_line((s)->outgoing, l))reconnect(NULL, s);}
#define CLIENT_SEND_LINE(c,l) {if(!irc_send_line((c)->incoming, l))disconnect_client(c);}

GList *servers = NULL;

extern FILE *debugfd;

extern char *debugfile;

void (*replicate_function) (struct network *, struct transport_context *client);

void data_hook_marshaller(GHook *h, gpointer data) {
	struct line *l = (struct line *)data;
	GHookFunc f;
	f = h->func;
	if(!(l->options & LINE_STOP_PROCESSING))f(data);
}

static void clients_send(struct network *server, struct line *l, struct transport_context *exception);
static void reconnect(struct transport_context *c, void *_server);

void handle_server_receive (struct transport_context *c, char *raw, void *_server)
{
	struct network *server = (struct network *)_server;
	char *new_nick;
	struct line *l;
	
	if(debugfd)fprintf(debugfd, "[From server] %s\n", raw);

	l = irc_parse_line(raw);
	
	l->direction = FROM_SERVER;
	l->network = server;

	state_handle_data(server,l);

	g_hook_list_marshal(&data_hook, FALSE, data_hook_marshaller, l);

	/* We need to handle pings.. we can't depend on a client
	 *          * to do that for us*/
	l->direction = FROM_SERVER;
	if(l->args[0]) {
		if(!strcasecmp(l->args[0], "PING")){
			if(server->outgoing)irc_send_args(server->outgoing, "PONG", l->args[1], NULL);
		} else if(!strcasecmp(l->args[0], "PONG")){
		} else if(!strcasecmp(l->args[0], "433") && !server->authenticated){
			if(server->outgoing) {
				char *nick = xmlGetProp(server->xmlConf, "nick");
				asprintf(&new_nick, "%s_", nick);
				xmlSetProp(server->xmlConf, "nick", new_nick);
				irc_send_args(server->outgoing, "NICK", new_nick, NULL);
				xmlFree(nick);
			}
		} else if(server->authenticated) {
			if(!(l->options & LINE_DONT_SEND))clients_send(server, l, NULL);
		} else if(atoi(l->args[0]) == 4) {
			xmlNodePtr cur = server->xmlConf->xmlChildrenNode;
			server->authenticated = 1;
			while(cur) {
				if(!strcmp(cur->name, "autosend")) {
					xmlChar *data = xmlNodeGetContent(cur);
					struct line *newl;
					
					newl = irc_parse_line(data);

					SERVER_SEND_LINE(server, newl);

					xmlFree(data);
					free_line(newl);
				}
				cur = cur->next;
			}
			
			/* Rejoin channels */
			cur = server->xmlConf->xmlChildrenNode;
			while(cur) {
				if(!strcmp(cur->name, "channel")) {
					char *n = xmlGetProp(cur, "name");
					char *a = xmlGetProp(cur, "autojoin");
					g_assert(n);
					if(a && atoi(a))
						irc_send_args(server->outgoing, "JOIN", n, NULL);
					if(n)xmlFree(n);
					if(a)xmlFree(a);
				} 
				cur = cur->next;
			}
		}
	}
	free_line(l); l = NULL;
}



gboolean login_server(struct network *s) {
	char *server_name = xmlGetProp(s->xmlConf, "name"),
	     *fullname, *nick, *username, *password;
	
	if(!s->current_server){ 
		xmlSetProp(s->xmlConf, "autoconnect", "0");
		g_warning("No servers listed for network %s, aborting reconnect.\n", server_name);
		xmlFree(server_name);
		return FALSE;
	}

	s->outgoing = transport_connect(s->current_server->name, s->current_server, handle_server_receive, reconnect, s);

	if(!xmlHasProp(s->xmlConf, "name") && xmlHasProp(s->current_server, "name")) {
		xmlFree(server_name);
		server_name = xmlGetProp(s->current_server, "name");
		xmlSetProp(s->xmlConf, "name", server_name);
	}

	if(!s->outgoing) {
		/* FIXME: handle comments */
		s->current_server = s->current_server->next;
		if(!s->current_server)s->current_server = s->servers->xmlChildrenNode;
		xmlFree(server_name);
		return FALSE;
	}


	nick = xmlGetProp(s->xmlConf, "nick");
	username = xmlGetProp(s->xmlConf, "username");
	fullname = xmlGetProp(s->xmlConf, "fullname");
	password = xmlGetProp(s->xmlConf, "password");
	g_assert(strlen(nick));
	g_assert(strlen(username));
	g_assert(strlen(fullname));
	g_assert(strlen(my_hostname));
	g_assert(strlen(server_name));
	
	if(xmlHasProp(s->xmlConf, "password"))irc_send_args(s->outgoing, "PASS", password, NULL);
	irc_send_args(s->outgoing, "NICK", nick, NULL);
	irc_send_args(s->outgoing, "USER", username, my_hostname, server_name, fullname, NULL);

	xmlFree(server_name);
	xmlFree(nick);
	xmlFree(username);
	xmlFree(fullname);
	xmlFree(password);

	return TRUE;
}

void reconnect(struct transport_context *c, void *_server)
{
	struct network *server = (struct network *)_server;
	char *server_name;
	/* Don't report disconnections twice */
	g_assert(server);
	server_name = xmlGetProp(server->xmlConf, "name");
	
	if(!server->outgoing) return;
	transport_free(server->outgoing); server->outgoing = NULL;

	g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Connection to server %s lost, trying to reconnect...", server_name);
	xmlFree(server_name);

	server->authenticated = 0;
	state_reconnect(server);
	g_timeout_add(RECONNECT_INTERVAL, (GSourceFunc) login_server, server);
}

void disconnect_client(struct client *c) {
	if(!c->incoming)return;

	transport_free(c->incoming);
	c->incoming = NULL;

	c->network->clients = g_list_remove(c->network->clients, c);
	c->authenticated = 2;
	free(c);
}

static void clients_send(struct network *server, struct line *l, struct transport_context *exception) 
{
	GList *g = server->clients;
	while(g) {
		struct client *c = (struct client *)g->data;
		if(c->authenticated && c->incoming != exception)CLIENT_SEND_LINE(c, l);
		g = g_list_next(g);
	}
}

void handle_client_disconnect(struct transport_context *c, void *data)
{
	struct client *client = (struct client *)data;
	disconnect_client(client);
}

void handle_client_receive(struct transport_context *c, char *raw, void *_client)
{
	struct client *client = (struct client *)_client;
	char allmodes[] = "aAbBcCdDeEfFgGhHiIjJkKlLmMnNoOpPqQrRsStTuUvVwWxXyYzZ";
	struct line *l;
	char *clientpass;
	char *server_name, *nick;

	if(client->authenticated == 2)return;
	
	if(debugfd)fprintf(debugfd, "[From client] %s\n", raw);

	l = irc_parse_line(raw);
	
	l->direction = TO_SERVER;
	l->client = client;
	l->network = client->network;

	state_handle_data(client->network, l);
	
	g_hook_list_marshal(&data_hook, FALSE, data_hook_marshaller, l);

	clientpass = xmlGetProp(client->network->xmlConf, "client_pass");
	if(!l->args[0]){ 
		free_line(l);
		return;
	}

	server_name = xmlGetProp(client->network->xmlConf, "name");
	nick = xmlGetProp(client->network->xmlConf, "nick");

	if(!clientpass && !client->authenticated) {
		client->authenticated = 1;
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "No password set for %s, allowing client _without_ authentication!", server_name);
	}

	if(!strcasecmp(l->args[0], "USER")){
		if(!client->authenticated){
			irc_sendf(c, ":%s 451 %s :You are not registered\r\n", 
					  server_name,
					  nick);
			free_line(l);l = NULL;
			if(clientpass)xmlFree(clientpass);
			xmlFree(server_name); xmlFree(nick);
			return;
		}
		irc_sendf(c, ":%s 001 %s :Welcome to the ctrlproxy\r\n", server_name, nick);
		irc_sendf(c, ":%s 002 %s :Host %s is running ctrlproxy\r\n", server_name, nick, my_hostname);
		irc_sendf(c, ":%s 003 %s :Ctrlproxy (c) 2002 Jelmer Vernooij <jelmer@nl.linux.org>\r\n", server_name, nick);
		irc_sendf(c, ":%s 004 %s %s %s %s %s\r\n", 
				server_name, nick, server_name, PACKAGE_VERSION, allmodes, allmodes);
		irc_sendf(c, ":%s 422 %s :No MOTD file\r\n", 
				server_name, nick);
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_MESSAGE, "Client @%s successfully authenticated", 
			  server_name);

		replicate_function(client->network, c);
	} else if(!strcasecmp(l->args[0], "PASS")) {
		if (!clientpass)
			client->authenticated = 1;
		else if(!strcmp(l->args[1], clientpass)) {
			client->authenticated = 1;
		} else {
			g_warning("User tried to log in to %s with incorrect password!\n", server_name);
			irc_sendf(c, ":%s 464 %s :Password mismatch\r\n", server_name, nick);
			disconnect_client(client);
			free_line(l);
			if(clientpass) xmlFree(clientpass);
			xmlFree(nick); xmlFree(server_name);
			return;
		}
	} else if(!strcasecmp(l->args[0], "NICK") && l->args[1] && !strcasecmp(l->args[1], nick)) {
		/* Ignore attempts to change nick to the current nick */
	} else if(!strcasecmp(l->args[0], "QUIT")) {
		disconnect_client(client);
		free_line(l);
		if(clientpass)xmlFree(clientpass);
		return;
	} else if(!strcasecmp(l->args[0], "PING")) {
		irc_sendf(c, ":%s PONG :%s\r\n", server_name, l->args[1]);
	} else if(client->authenticated) {
		if(client->network->outgoing) {
			const char *old_origin;
			if(!(l->options & LINE_DONT_SEND)) {SERVER_SEND_LINE(client->network, l) }

			/* Also write this message to all other clients currently connected */
			if(!(l->options & LINE_IS_PRIVATE)) {
				old_origin = l->origin; l->origin = nick;
				clients_send(client->network, l, c);
				l->origin = old_origin;
			}
		} else {
			irc_sendf(c, ":%s NOTICE %s :Currently not connected to server, ignoring\r\n", (server_name != NULL?server_name:"ctrlproxy"), nick);
		}
	} else {
		irc_sendf(c, ":%s NOTICE %s :You are not logged in yet. Ignoring\r\n", server_name?server_name:"ctrlproxy", nick);
	}
	free_line(l);l = NULL;
	if(clientpass)xmlFree(clientpass);
	xmlFree(server_name); xmlFree(nick);
}

void handle_new_client(struct transport_context *c_server, struct transport_context *c_client, void *_server)
{
	struct network *n = (struct network *)_server;
	struct client *d = malloc(sizeof(struct client));

	d->authenticated = 0;
	d->network = n;
	d->connect_time = time(NULL);
	d->incoming = c_client;

	n->clients = g_list_append(n->clients, d);

	transport_set_receive_handler(c_client, handle_client_receive);
	transport_set_disconnect_handler(c_client, handle_client_disconnect);
	transport_set_data(c_client, d);
}

struct network *connect_to_server(xmlNodePtr conf) {
	uid_t uid;
	struct network *s = malloc(sizeof(struct network));
	char *nick, *user_name;
	xmlNodePtr cur;

	memset(s, 0, sizeof(struct network));
	s->xmlConf = conf;
	if(!xmlHasProp(s->xmlConf, "nick"))
		xmlSetProp(s->xmlConf, "nick", g_get_user_name());

	uid = getuid();

    if(!xmlHasProp(s->xmlConf, "username"))xmlSetProp(s->xmlConf, "username", g_get_user_name());

   if(!xmlHasProp(s->xmlConf, "fullname"))xmlSetProp(s->xmlConf, "fullname", g_get_real_name());
	
   	nick = xmlGetProp(s->xmlConf, "nick");
	user_name = xmlGetProp(s->xmlConf, "username");
	asprintf(&s->hostmask, "%s!~%s@%s", nick, user_name, my_hostname);
	xmlFree(nick); xmlFree(user_name);

	s->listen = transport_listen("ipv4", s->xmlConf, handle_new_client, s);
	if(!s->listen) {
		g_log(G_LOG_DOMAIN, G_LOG_LEVEL_WARNING, "Can't listen!");
	}

	s->servers = NULL;
	s->current_server = NULL;
	
	/* Find <servers> tag */
	cur = s->xmlConf->xmlChildrenNode;
	while(cur) {
		if(!strcmp(cur->name, "servers"))s->servers = cur;
		cur = cur->next;	
	}

	if(!s->servers) {
		char *server_name = xmlGetProp(s->xmlConf, "name");
		g_warning("No servers listed for network %s!\n", server_name);
		xmlFree(server_name);
	} else {
		s->current_server = s->servers->xmlChildrenNode;	
	}
	
	/* Add server by default. If connecting succeeds, it is 
	 * removed automagically by login_server */
	login_server(s);
	servers = g_list_append(servers, s);

	return s;
}

int close_server(struct network *s)
{
	GList *l = s->clients;
	char *server_name = xmlGetProp(s->xmlConf, "name");
	if(s->outgoing)irc_sendf(s->outgoing, ":%s QUIT\r\n", server_name);
	free(s->hostmask);
	
	while(l) {
		struct client *c = l->data;
		irc_sendf(c->incoming, ":%s QUIT :Server exiting\r\n", server_name);
		transport_free(c->incoming);
		l = l->next;
	}
	g_list_free(s->clients);s->clients = NULL;
	transport_free(s->listen);

	transport_free(s->outgoing);

	servers = g_list_remove(servers, s);

	xmlFree(server_name);
	return 0;
}


gboolean ping_loop(gpointer user_data) {
	GList *l = servers;
	while(l) {
		struct network *s = (struct network *)l->data;
		GList *cl;
		char *server_name = xmlGetProp(s->xmlConf, "name");
		if(s->outgoing)irc_send_args(s->outgoing, "PING", server_name, NULL);
		else reconnect(NULL, s);
		xmlFree(server_name);
		
		/* Throw out unauthorized clients that have been inactive for over a minute */
		cl = s->clients;
		while(cl) {
			struct client *c = (struct client *)cl->data;
			cl = g_list_next(cl);
			if(c->authenticated)continue;
			if(time(NULL) - c->connect_time > 60) {
				disconnect_client(c);
			}
		}
		
		l = g_list_next(l);
	}
	return TRUE;
}

void default_replicate_function(struct network *s, struct transport_context *io) {
	GSList *g = gen_replication(s);
	while(g) {
		struct line *rl = (struct line *)g->data;
		if(rl) {
			irc_send_line(io, rl);
			free_line(rl);
		}
		g = g_slist_next(g);
	}
	g_slist_free(g);
}
